/*
 * Copyright (c) 1995 by CompuServe Incorporated, Columbus, Ohio
 *
 * Facility:  libinetcis.a
 *
 * Abstract:
 *
 *     This module contains the I/O operations that can be performed on a
 *     socket in inetcis programs.
 *
 * Environment:  BSDI's BSD/OS 1.1 (POSIX)
 *
 * Author:  William L. Chatfield (BCHATFIELD), 14-Jul-95
 *
 * Revision History:
 *
 * 1.0  William L. Chatfield (BCHATFIELD), 14-Jul-95
 *      - Initial revision
 */

#ifndef SOCKIO_H
#define SOCKIO_H

/*
 * External Interfaces
 */
#include <comdef.h>                  /* CompuServe definitions */

/*
 * Constants, structures, and typedefs
 */

/*
 * External data
 */

/*
 * Function prototypes
 */

int Write_To_Socket (int socket_file_descriptor, char *buffer_ptr,
		     int number_of_bytes_to_write, int *number_written);
/*
 * Functional Description:
 *
 *      This function writes to a socket and handles any errors.
 *
 * Formal Parameters:
 *
 *      socket_file_descriptor (Read only)
 *          - The descriptor of the socket to which to write.
 *      buffer_ptr (Write only)
 *          - Pointer to buffer to which data will be taken.
 *      number_of_bytes_to_write (Read only)
 *          - Number of bytes to write.
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      The number of bytes written is returned.
 *
 * Side Effects:  None
 *
 */

int Read_From_Socket (int socket_file_descriptor, char *buffer_ptr,
		      int number_of_bytes_to_read, int *number_read);
/*
 * Functional Description:
 *
 *      This function reads from a socket and handles any errors and EOF
 *      contitions.
 *
 * Formal Parameters:
 *
 *      socket_file_descriptor (Read only)
 *          - The socket's file descriptor.
 *      buffer_ptr (Read only)
 *          - Pointer to the buffer in which data is put.
 *      number_of_bytes_to_read (Read only)
 *          - Number of bytes to read.
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      The number of bytes read is returned.
 *
 * Side Effects:  None
 *
 */

int Write_All_To_Socket (int file_descriptor, char *buffer_ptr,
			 int number_of_bytes_to_write);
/*
 * Functional Description:
 *
 *      This function writes all the bytes specified and does not return until
 *      it has written all of them or an error occurs.
 *
 * Formal Parameters:
 *
 *      file_descriptor (Read only)
 *          - descriptor to write to
 *      buffer_ptr (Read only)
 *          - pointer to the buffer from which the bytes written will be taken
 *      number_of_bytes_to_write (Read only)
 *          - number of bytes to write
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      SUCCESS, FAILURE
 *
 * Side Effects:  None
 *
 */

void Enable_Socket_Inactivity_Timeout (int sockdes);
/*
 * Functional Description:
 *
 *      This function turns on the SO_KEEPALIVE socket option, which requires
 *      the remote host to respond periodic "pings" from local host or the
 *      connection is broken.
 *
 * Formal Parameters:
 *
 *      sockdes (Read only)
 *          - The descriptor for the socket to be used.
 *
 * Implicit Parameters:  None
 *
 * Return Value:
 *
 *      None
 *
 * Side Effects:  Exit on error.  Be sure you have an exit handler.
 *
 */

/* It is possible Convert_h_errno_To_Text is not needed if hstrerrno
   is availiable. */

int Convert_h_errno_To_Text (int h_errno, char *h_errno_text);
int Write_All_To_Socket_EonF (int fd, char *buffer, int howmany);
void Log_Socket_Stats (); void Start_Producing_Socket_Flow_Stats ();
void Purge_Socket_Input (int sock_fd);

#endif
