/*
 * Copyright (c) 1992 by CompuServe Incorporated, Columbus, Ohio
 *
 * $Header: /u3/bchatfie/cvsroot/inetcis/inetcis.login/inetcis.login.c,v 1.2 1996/03/26 00:35:15 bchatfie Exp $
 *
 * Facility:
 *	Terminal emulator
 *
 * Abstract:
 *	Quick and dirty terminal emulator for Unix.  Simply moves
 *	characters from TTY to the device entered on the command
 *	line.  Three ESC's in a row exits.
 *
 * Environment:
 *
 *	BSD Unix.
 *
 * Author:
 *
 *	David J. Crone, Systems/Monitor Group.
 *
 * Revision History:
 *	January 6, 1992		DJC	Creation
 *      July 1, 1993            BBIGGS  Co-erced into inetcis throughput
 *         - speed throughput enforcement
 *         - disabled 3 escape exitor
 *         - determine tpad from tty
 *      January 27, 1994        BCHATFIELD Removed unavailable items
 *         - removed 14.4 baud option from prompt
 *         - removed item 5 from help menu
 *         - changed wording of sign-up prompt
 *
 * $Log: inetcis.login.c,v $
 * Revision 1.2  1996/03/26  00:35:15  bchatfie
 * There should be no modifications for inetcis.login.
 *
 * Revision 1.1.1.1  1995/08/24  14:30:02  bchatfie
 * Import of inetcis as one big module.
 *
 * Revision 1.2  1995/06/20  20:20:02  bchatfie
 * Changed error message for when no X.25 device is not found
 * to something more useful.
 *
 * Revision 1.1.1.1  1995/04/25  17:47:50  bchatfie
 * Imported sources
 *
 * Revision 1.6  1994/06/01  16:06:56  bchatfie
 * Temporary fix to handle text to binary mode switching problem.
 * Extra NULLs are removed from the beginning of strings in
 * trimlower.
 *
 * Revision 1.5  1994/05/16  17:07:51  bchatfie
 * Fixed Host Name: prompt help code.  Added code to read and display file
 * inetcis.text.host_name_help when incorrect input is read at the Host Name:
 * prompt.
 *
 * Revision 1.4  1994/05/05  15:54:13  bchatfie
 * Entered code to allow it to accept baud rates of 450, 1200, 1800, 4800,
 * 14400, 19200, and 38400 in addition to the ones it already accepts.
 * Also removed the extra line feed before the Host Name: prompt.
 *
 * Revision 1.3  1994/05/05  13:33:07  bchatfie
 * Fixed text directory location for test version.
 *
 * Revision 1.2  1994/05/05  10:16:59  bchatfie
 * I inserted some carriage returns.
 *
 * Revision 0.11  1994/03/30  16:19:50  bchatfie
 * Fixed line feed problem at the User ID prompt by turning on ONLCR in the
 * c_oflag field of the DevTermios structure.
 *
 * Revision 0.10  1994/03/29  18:24:43  bchatfie
 * This version does Control-D disconnects, User ID prompt timeout, and has
 * the beta test password removed.  It is the first version released to the
 * public.
 *
 * Revision 0.9  1994/03/23  09:00:07  bchatfie
 * This revision includes all the changes that have been made to the text and
 * prompts that are displayed for the user before he gets the User ID prompt.
 *
 * Revision 0.8  1993/12/05  13:43:38  bbiggs
 * fix to fgets to handle the "EOF" response properly (exits)
 *
 * Revision 0.7  1993/11/30  02:25:09  bbiggs
 * added 2 minute timeout at Host Name prompt
 * took out debug message at login
 * fixed typo in one of the text files (you electronix to your electronic)
 *
 * Revision 0.6  1993/10/22  21:33:35  bbiggs
 * Added HELP to the FIRST Host Name prompt. Changed Bit Rate to Baud Rate
 * in error message.
 *
 * Revision 0.5  1993/10/22  20:44:58  bbiggs
 * Changes for HELP in the baud rate prompt, and at the Host Name prompt
 *
 * Revision 0.3  1993/10/14  13:31:50  bbiggs
 * Changes to the flow of the user interrogation sequences
 * added speed 14400 to the list of available baud rates
 * added syslog messages for connect attempts and mailings
 *
 * $Log: inetcis.login.c,v $
 * Revision 1.2  1996/03/26  00:35:15  bchatfie
 * There should be no modifications for inetcis.login.
 *
 * Revision 1.1.1.1  1995/08/24  14:30:02  bchatfie
 * Import of inetcis as one big module.
 *
 * Revision 1.2  1995/06/20  20:20:02  bchatfie
 * Changed error message for when no X.25 device is not found
 * to something more useful.
 *
 * Revision 1.1.1.1  1995/04/25  17:47:50  bchatfie
 * Imported sources
 *
 * Revision 1.6  1994/06/01  16:06:56  bchatfie
 * Temporary fix to handle text to binary mode switching problem.
 * Extra NULLs are removed from the beginning of strings in
 * trimlower.
 *
 * Revision 1.5  1994/05/16  17:07:51  bchatfie
 * Fixed Host Name: prompt help code.  Added code to read and display file
 * inetcis.text.host_name_help when incorrect input is read at the Host Name:
 * prompt.
 *
 * Revision 1.4  1994/05/05  15:54:13  bchatfie
 * Entered code to allow it to accept baud rates of 450, 1200, 1800, 4800,
 * 14400, 19200, and 38400 in addition to the ones it already accepts.
 * Also removed the extra line feed before the Host Name: prompt.
 *
 * Revision 1.3  1994/05/05  13:33:07  bchatfie
 * Fixed text directory location for test version.
 *
 * Revision 1.2  1994/05/05  10:16:59  bchatfie
 * I inserted some carriage returns.
 *
 * Revision 0.11  1994/03/30  16:19:50  bchatfie
 * Fixed line feed problem at the User ID prompt by turning on ONLCR in the
 * c_oflag field of the DevTermios structure.
 *
 * Revision 0.10  1994/03/29  18:24:43  bchatfie
 * This version does Control-D disconnects, User ID prompt timeout, and has
 * the beta test password removed.  It is the first version released to the
 * public.
 *
 * Revision 0.9  1994/03/23  09:00:07  bchatfie
 * This revision includes all the changes that have been made to the text and
 * prompts that are displayed for the user before he gets the User ID prompt.
 *
 * Revision 0.8  1993/12/05  13:43:38  bbiggs
 * fix to fgets to handle the "EOF" response properly (exits)
 *
 * Revision 0.7  1993/11/30  02:25:09  bbiggs
 * added 2 minute timeout at Host Name prompt
 * took out debug message at login
 * fixed typo in one of the text files (you electronix to your electronic)
 *
 * Revision 0.6  1993/10/22  21:33:35  bbiggs
 * Added HELP to the FIRST Host Name prompt. Changed Bit Rate to Baud Rate
 * in error message.
 *
 * Revision 0.5  1993/10/22  20:44:58  bbiggs
 * Changes for HELP in the baud rate prompt, and at the Host Name prompt
 *
 * Revision 0.2  1993/10/14  13:30:51  bbiggs
 * Various changes of formatting. Some clean-up to the throughput
 * logic.
 *
 * Revision 0.1  1993/09/07  14:47:58  bbiggs
 * Initial checkin of software. Preceeds addition of the exponential
 * averaging enforcement of the speed throughput.
 */

/*
 * External interfaces
 */
#include <sys/types.h>  /* for struct stat, ... */
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>   /* for stat() */
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>     /* for getppid */
#include <termios.h>    /* for B9600 */
#include <signal.h>     /* for kill */
#include <setjmp.h>
#include <ctype.h>
#include <syslog.h>
#include <string.h>     /* for strcpy(), strcat(), ... */
#include <limits.h>     /* for INT_MAX */
#include <comdef.h>
#include "config.h"     /* for LOCATION_OF_TEXT_FILES */
#include "../common/get_tpad.h"  /* for get_tpad() */
#include "../common/inetcis_common.h"

#define STRING__BUFFER_SIZE              256
#define STRING__MAXIMUM_LENGTH           (STRING__BUFFER_SIZE - 1)
#define STRING__FIRST_POSITION_IN_BUFFER 0
#define STRING__LAST_POSITION_IN_BUFFER  (STRING__BUFFER_SIZE - 1)

const char *PROGRAM_ID_FOR_ERRORS = "ILG";
const char *OUTPUT_LINE_TERMINATOR = "\n";
const double TIMEOUT_VALUE = 30;  /* seconds */

/*
 * Function prototypes
 */
int Initial_Tty_SetUp ();
int Initial_Connect (char *, char *);
int Get_Password ();
int Get_Service ();
int Get_First_Prompt ();
int Get_Agreement ();
int Get_Help ();
int Perform_Dialogs ();
int SetUp_Devices (int *, int *, int);
void dumpit (FILE *, char *, int);
void Display_Help_Menu ();
void Display_Text_File (char *);
void Strip_Leading_CRs_And_LFs (char *, char *, int, int *);
void Read_Asterisk_Prompt ();
void Alarm_Signal_Handler (int);
void Write_All_Characters (FILE *, char *, char *, int);
void exit_handler ();
void Get_Baud_Rate_from_User (int *, int *, int *);
void terminate_handler (int);

#define BUFSIZE         4096+1
#define HOSTNAME_BAD    0
#define HOSTNAME_CIS    1
#define HOSTNAME_CPS    2
#define HOSTNAME_AGREE  3

/* If these values are not lowercased they will not match! */
char *HostNameList[] = {"cis", "cps", "cisagree", (char *)NULL};

#define GetServiceText  "inetcis.text.host_name_help"
#define HostNamePrompt "\r\rHost Name: "

/* Values associated with Get_Help() routine */

#define GetHelpText1    "inetcis.text.6a"
#define GetHelpText2    "inetcis.text.6b"
#define GetHelpText3    "inetcis.text.7"
#define GetHelpText4    "inetcis.text.8"

#define GetPasswordText "inetcis.text.1"
#define GetPasswordGood "inetcis.text.1a"
#define GetPasswordBad  "inetcis.text.1b"

#define GetFirstPrompt1 "inetcis.text.2a"
#define GetFirstPrompt2 "inetcis.text.2"

#define GetAgreeText1   "inetcis.text.3a"
#define GetAgreeText2   "inetcis.text.3b"
#define GetAgreeText3   "inetcis.text.3c"

#define ExplainBitRate  "inetcis.text.4"

#define GetHelpPrompt1  \
        "Enter choice (blank line, number(s) between 1 and 6, OFF): "

#define show_version \
    printf ("InetCIS ALPHA 1.0 %s11/15/93\n", logstr); fflush (stdout);

#define showoriginip \
    printf ("ORIGINIP [%s] device[%s]\r\n", getenv("ORIGINIP"), ttyname(0))

int MaxFd = -1;                 /* max of DevFd, TtyFd (used for select call)*/
int DevFd = -1;                 /* Device File descriptor */
int TtyFd = -1;                 /* TTY File descriptor */
int DevCc = 0;                  /* Device input buffer char count */
int TtyCc = 0;                  /* TTY input buffer char count */
char *DevBufPtr;                /* Device buffer pointer */
char *TtyBufPtr;                /* TTY buffer pointer */
char DevBuf[BUFSIZE];           /* Device input buffer */
char TtyBuf[BUFSIZE];           /* TTY input buffer */
char IP_Address[STRING__BUFFER_SIZE];


char DevName[128];              /* ptr to device name */
char oDevName[128];             /* ptr to device name */

struct termios TtyTermios;      /* working copy */
struct termios TtySettings;     /* original settings */
struct termios DevTermios;      /* working copy */
struct termios DevSettings;     /* original settings */

#ifdef log
char *logstr = "(debug log) ";
#else
char *logstr = "";
#endif


jmp_buf asterisk_read_timeout;

#if defined(log) || defined(TEST_LEVEL)
     FILE *debugfile;
#endif


/*
 * Functional Description:
 *      main
 *
 * Formal Parameters:
 *      argc
 *      argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

void main (int argc, char *argv[])
{
    char stripped_buffer[BUFSIZE];
    int stripped_buffer_length = 0;
    int character_count;
    fd_set ibits, obits, xbits;
    int nsel;
    char ch;
    int cc;
    int CharsIn, CharsOut, junk;
    int MaxChar=0, BaudRate=0, ReadWait;
    int HnResult;
    time_t old_time, current_time;
    struct timeval tv_s, delay;
    int DebugPrint = 0;
    int ExpRate1, ExpRate2;
    void (*previous_signal_handler) (int);
    int system_status;


    /*
     * Open the log file.
     */
    openlog ("inetcis.login", LOG_PID, LOG_LOCAL7);
    syslog (LOG_DEBUG, "log opened");

    /* register exit handler */
    system_status = atexit (exit_handler);
    if (system_status == ATEXIT_ERROR)
        {
        syslog (LOG_ERR, "%s %d - err - function atexit caused system error"
                " `%m' trying to define exit handler exit_handler",
		__FILE__, __LINE__);
	}

    /*
     * Specify interrupt handlers.
     */
    previous_signal_handler = signal (SIGTERM, terminate_handler);
    if (previous_signal_handler == SIG_ERR)
	{
	syslog (LOG_ERR, "before line %d signal caused: %m", __LINE__);
	exit (EXIT_FAILURE);
	}
    previous_signal_handler =  signal (SIGSEGV, segmentation_fault_handler);
    if (previous_signal_handler == SIG_ERR)
	{
	syslog (LOG_ERR, "before line %d signal caused: %m", __LINE__);
	exit (EXIT_FAILURE);
	}

#if defined(log) || defined(TEST_LEVEL)
    char dbgfilename[128];
    sprintf (dbgfilename, "/home/bchatfield/inetcis.%d.log", getpid());
    debugfile = fopen (dbgfilename, "w");
#endif

    /*
      * Process command line arguments.
     */
    while ((ch = getopt (argc, argv, "fpxd:i:")) != EOF)
	{
	switch (ch)
	    {
	    case 'i':
                strncpy (IP_Address, optarg, STRING__MAXIMUM_LENGTH);
		IP_Address[STRING__LAST_POSITION_IN_BUFFER] = EOS;
                break;

	    /*
	     * part of the dummied up login from telnetd passes
	     * in these arguments. Ignore them.
	     */
	    case 'f':
	    case 'p':
		break;

	    case 'd':
		strcpy (DevName, optarg);
		break;

	    case 'x':
		DebugPrint = 1;
		break;

	    case '?':
	    case 'h':
	    default:
		printf ("usage: inetcis.login"
			" -d device_name"
			" -i ip_address"
			"\n");
		fflush (stdout);
		sleep (2);
		exit (1);
		break;
	    }  /* switch */
	}  /* while */

    /*
     * Log IP address of caller.
     */
    syslog (LOG_INFO, "connect from %s", IP_Address);


    /*
     * Before we do anything, lets not assume the terminal device state
     * is anywhere near normal... fix it up for what we expect it to be!
     */
    Initial_Tty_SetUp ();
    
#ifdef log
    show_version;
    fprintf (debugfile, "Device name \"%s\"\n", DevName);
    fflush (debugfile);
    printf ("Device name \"%s\"\n", DevName);
    fflush (stdout);
#endif

#if 0
    /* Debug banner and Device name mapping */
    Initial_Connect (DevName, oDevName);
#endif

    /*
     * Display the "Connected to CompuServe" message.
     */
    Display_Text_File (GetPasswordText);

#if 0
    /*
     * This was for ONLY the betatest period, which ended March 30, 1994.
     */
    if (!Get_Password ())
	{
	sleep (1);
	exit (EXIT_SUCCESS);
	}
#endif

    /*
     * Simulate the "Host Name:" prompt.
     */
    HnResult = Get_Service();
    if (HnResult == HOSTNAME_BAD)
	{
	sleep (1);
	exit (EXIT_SUCCESS);  /* EXIT_SUCCESS? */
	}
    else if (HnResult == HOSTNAME_CIS || HnResult == HOSTNAME_CPS)
    	{
	/* We might not return from this! */
	Perform_Dialogs ();
	}

    /*
     * All baud rate information is hard-coded and there is no throttling
     * done anymore, so users don't want to see a baud rate prompt anymore.
     */
#if 0
    Get_Baud_Rate_from_User (&Baud, &BaudRate, &MaxChar);
#endif
    BaudRate = B57600;
    MaxChar = INT_MAX;

    /*
     * Bring up the devices, raw mode, one to the service and
     * one back to the user. (service == X.25, user == tcp/ip)
     */
    if (!SetUp_Devices (&DevFd, &TtyFd, BaudRate))
	{
	syslog (LOG_ERR, "Unable to open any X.25 device.");
	exit (EXIT_FAILURE);
	}
    syslog (LOG_INFO, "using X.25 device %s", DevName);

    MaxFd = (TtyFd > DevFd) ? TtyFd + 1 : DevFd + 1;

    /*
     * main loop -
     * Sit here and shuffle characters back and forth
     */

    CharsIn = 0;
    CharsOut = 0;
    old_time = time (NULL);
    current_time = time (NULL);

    /*
     * When a pad device is opened, it responds with an asterisk
     * prompt. We need to wait for it to arrive before sending
     * our connect request out.
     */

    Read_Asterisk_Prompt ();

#if 0
    DevCc = read (DevFd, DevBuf, BUFSIZE);
    DevBufPtr = DevBuf;
    if (DebugPrint)
	{
#ifdef log
	fprintf (debugfile, "\r\nrdev: (%d) ", DevCc);
#endif
	printf ("\r\nrdev: (%d) ", DevCc);
	dumpit (stdout, DevBuf, DevCc);
	}

    if (DevCc != 3 || strcmp ("\015\012*", DevBuf) != 0)
	{
#ifdef log
	fprintf (debugfile, "Connection attempt to CIS failed.\r\n");
#endif
	printf ("Connection attempt to CIS failed.\r\n");
	fflush (stdout);
	exit (0);
	}
#endif

    /*
     * Set the CompuServe extended parameter 45 to 255.  This can be done now
     * that the asterisk prompt has been received from the X.25 device.
     */
    strcpy (TtyBuf, "xset 45:255\r");
    TtyCc = strlen (TtyBuf);
    character_count = write (DevFd, TtyBuf, TtyCc);
    if (character_count < 0)
	{
	syslog (LOG_ERR, "before line %d write of xset 45:255 caused: %m",
		__LINE__);
	}

    /*
     * If this call gets an error it will probably not return.
     */
    Read_Asterisk_Prompt ();

    /*
     * Set up the call the information service
     * Use of 313299999997 will get the user to the
     * "User ID" prompt, and lock them to that point.
     * To release this, (e.g. to access the host name
     * prompt, change the address to just 3132
     */
    strcpy (TtyBuf, "313299999997\r");
    TtyCc = strlen (TtyBuf);

    cc = write (DevFd, TtyBuf, TtyCc);
    if (DebugPrint)
	{
#ifdef log
	fprintf (debugfile, "wdev: (%d) ", cc);
#endif
	printf ("wdev: (%d) ", cc);
	dumpit (stdout, TtyBuf, TtyCc);
	}

    /*
     * Pick up the echoed connect request...
     */
    DevCc = read (DevFd, DevBuf, BUFSIZE);
    Strip_Leading_CRs_And_LFs (DevBuf, stripped_buffer, DevCc,
			       &stripped_buffer_length);
    DevBufPtr = DevBuf;
    if (DebugPrint)
	{
#ifdef log
	fprintf (debugfile,"\r\nrdev: (%d) ", DevCc);
	dumpit (debugfile, DevBuf, DevCc);
#endif
	printf ("\r\nrdev: (%d) ", DevCc);
	dumpit (stdout, DevBuf, DevCc);
	}

    /*
     * A bunch of pad messages get sent back on
     * the connection attempt. Rather than parse through
     * them, we just wait for our confirmation prompt, and
     * proceed from there.
     */

    read_user_id_prompt (DevFd);

    /*
     * The User ID prompt has successfully been read from the X.25 device.
     * Display it for the user.
     */

    printf ("\r\nUser ID: ");
    fflush (stdout);

    DevBuf[0] = '\0';
    TtyBuf[0] = '\0';
    DevCc = 0;
    TtyCc = 0;
    DevBufPtr = DevBuf;

    /* Initialize my select wait period... */
#ifdef HARD
    CharsIn = 0;
    CharsOut = 0;
#if 1
    tv_s.tv_sec = 1;
    tv_s.tv_usec = 0;
    ReadWait = 250;
#else
    tv_s.tv_sec = 0;
    tv_s.tv_usec = 250;
    MaxChar = MaxChar / 4;
    ReadWait = 250;
#endif
#else
    CharsIn = 0;
    CharsOut = 0;
    tv_s.tv_sec = 0;
    tv_s.tv_usec = 250;		/* milliseconds */
    MaxChar = MaxChar / 4;
    ReadWait = 250;
#endif

    ExpRate1 = 0;
    ExpRate2 = MaxChar;

    for (;;)
	{
	FD_ZERO (&ibits);
	FD_ZERO (&obits);
	FD_ZERO (&xbits);

#ifdef HARD
	/*
	 * NOTE: Slippage logic is not compiled in right now, just
	 * limit the user to the quantum bandwidth promised, and no
	 * more. Pending further testing...
	 */

	/*
	 * Begin slippage logic:
	 * If, during the preceeding time quantum, we were not able
	 * to use all of our alotted "bandwidth", allow the amount of
	 * "bandwidth" this quantum to increase by the unused portion.
	 * However, the increase should not exceed 2x the available
	 * quantum size.
	 */

	/*
	 * time is in increments of 1 second. That will be our quantum,
	 * strictly a convienience...
	 */
	current_time = time (NULL);

	if (current_time != old_time)
	    {
#if 0
	    /* The "slippage" logic is currently disabled here */
	    if (CharsIn < MaxChar)
	        CharsIn = 0 - (abs (CharsIn) > MaxChar ?
			       (Baud + 1) * abs (CharsIn) : MaxChar -
			       abs (CharsIn));
	    else
#endif
		CharsIn = 0;
#if 0
	    /* The "slippage" logic is currently disabled here */
	    if (CharsOut < MaxChar)
		CharsOut = 0 - (abs (CharsOut) > MaxChar ?
				(Baud + 1) * abs (CharsOut) : MaxChar -
				abs (CharsOut));
	    else
#endif
		CharsOut = 0;
	    old_time = current_time;
	    }

	/* End of slippage logic */
#endif
	/*
	 * Never look for input if there's still
	 * stuff in the corresponding output buffer
	 */

	if ((TtyCc >= 0) && (DevCc >= 0))
	    {			/* if tty & dev still alive */
	    if (DevCc)		/* if anything in device input buffer*/
		FD_SET (TtyFd, &obits);	/* then request tty output */
	    else if (CharsOut < MaxChar)
		FD_SET (DevFd, &ibits);	/* else request device input check */

	    if (TtyCc)		/* if anything in tty input buffer */
		FD_SET (DevFd, &obits);	/* then request device output */
	    else if (CharsIn < MaxChar)
		FD_SET (TtyFd, &ibits);	/* else request tty input check */
	    }

	if ((TtyCc < 0) || (DevCc < 0))	/* connection was closed - exit */
	    break;

	FD_SET (TtyFd, &xbits);	/* just for grins */
	FD_SET (DevFd, &xbits);

	/* check if any requested devices will block and reset
	   bits to those that are ready */

	nsel = select ((MaxFd + 1), &ibits, &obits, &xbits,
		       (struct timeval *) &tv_s);

/* Just pause here quietly for a moment - let the buffer fill up a bit */
	delay.tv_sec = 0;
	delay.tv_usec = ReadWait - tv_s.tv_usec;
	junk = select (0, NULL, NULL, NULL, (struct timeval *) &delay);

#ifndef HARD
	if (tv_s.tv_usec == 0 || nsel == 0)
	    {
	    tv_s.tv_sec = 0;
	    tv_s.tv_usec = ReadWait;
	    CharsIn = 0;
	    CharsOut = 0;
	    }
#endif
	/*
	 * Watch for the runaway condition... blow out if it's
         * encountered.
	 */

	if ( READ_ERROR == read(DevFd, DevBuf, 0) )
	    {
	    syslog (LOG_INFO, "before line %d read caused: %m", __LINE__);
	    break;
	    }

	if (READ_ERROR == read(TtyFd,TtyBuf,0))
	    {
	    syslog (LOG_INFO, "before line %d read caused: %m", __LINE__);
	    break;
	    }

	/*
	 * Something to read from the device...
	 * (nothing in buffer and read will not block)
	 */
	if ((DevCc == 0) && (FD_ISSET (DevFd, &ibits) && (MaxChar > CharsOut)))
	    {
	    DevCc = read (DevFd, DevBuf, MaxChar);
	    DevBufPtr = DevBuf;
	    CharsOut += DevCc;

#ifdef log
 	    fprintf (debugfile, "rdev: (%d) ", DevCc);
	    dumpit (debugfile, DevBuf, DevCc);
#endif
	    if (DevCc < 0)
		{
		if (errno == EWOULDBLOCK)
		    DevCc = 0;
		else
		    DevCc = -1;
		}
	    }

	/*
	 * Something to read from the tty...
	 */
	if ((TtyCc == 0) && FD_ISSET (TtyFd, &ibits) && (MaxChar > CharsIn))
	    {
	    TtyCc = read (TtyFd, TtyBuf, MaxChar);
	    TtyBufPtr = TtyBuf;
	    CharsIn += TtyCc;

#ifdef log
 	    fprintf (debugfile, "rtty: (%d) ", TtyCc);
	    dumpit (debugfile, TtyBuf, TtyCc);
#endif
	    if (TtyCc < 0)
		{
		if (errno == EWOULDBLOCK)
		    TtyCc = 0;
		else
		    TtyCc = -1;
		}
	    }

	/*
	 * Something to write to device
	 */
	if ((TtyCc > 0) && FD_ISSET (DevFd, &obits))
	    {
	    cc = write (DevFd, TtyBufPtr,
			((TtyCc > MaxChar) ? MaxChar : TtyCc));

#ifdef log
 	    fprintf (debugfile, "wdev: (%d) ", cc);
	    dumpit (debugfile, TtyBufPtr, cc);
#endif
	    if (cc >= 0)
		{
		TtyCc -= cc;
		TtyBufPtr += cc;
		}
	    else
		{
		/* log write error */
		}
	    }

	/*
	 * Something to write to tty
	 */
	if ((DevCc > 0) && FD_ISSET (TtyFd, &obits))
	    {
	    cc = write (TtyFd, DevBufPtr,
			(DevCc > MaxChar ? MaxChar : DevCc));
#ifdef log
 	    fprintf (debugfile, "wtty: (%d) ", cc);
	    dumpit (debugfile, DevBufPtr, cc);
#endif
	    if (cc >= 0)
		{
		DevCc -= cc;
		DevBufPtr += cc;
		}
	    else
		{
		/* log write error */
		}
	    }

	}			/* end of main loop */

    tcsetattr (TtyFd, TCSANOW, &TtySettings);	/* restore TTY settings */
    exit (0);

}

int Initial_Tty_SetUp ()
{
    int TtyFd;

#if 0

    /*
     * This variable is unused currently.
     */
    struct termios TtySettings;

#endif

    if ((TtyFd = open ("/dev/tty", O_RDWR)) < 0)
	{
	printf ("\nCould not open tty");
	fflush (stdout);
	return (0);
	}

    if (tcgetattr (TtyFd, &TtyTermios) < 0)
	{
	printf ("\nCould not get termios structure");
	fflush (stdout);
	return (0);
	}

    /*
     * ICANON - input processing
     * ISIG   - enable signals
     * TOSTOP - stop background prcesses if they write to the controlling term
     * ECHO   - echo typed characters
     */

    TtyTermios.c_iflag      |= (ICRNL | IXON | IXOFF);
    TtyTermios.c_oflag      |= (OPOST | ONLCR);
    TtyTermios.c_cflag      |= (CS7 | PARENB);
    TtyTermios.c_lflag      |= (ICANON | ISIG | TOSTOP | ECHO | EXTPROC);
    TtyTermios.c_lflag      &= ~ PENDIN;
    TtyTermios.c_cc[VERASE]  = '\010';

    if (tcsetattr (TtyFd, TCSANOW, &TtyTermios) < 0)
	{
	syslog (LOG_ERR, "before line %d tcsetattr caused: %m", __LINE__);
	return (0);
	}

    return (1);
}

/*
 * The contents of resp are not changed except that some of the characters may
 * be lower cased and it may be shorter.  Basically the contents of resp are
 * no longer any good.  The correct result is in the value that is returned,
 * which is a pointer to a substring of resp.  That result has the white space
 * trimmed off the front and back and has the alphanumeric characters lower
 * cased.  The way this thing invalidates the contents of resp is screwey, but
 * it may be dangerous to change this functionality.
 */
char *trimlower (char *resp, int len)
    {
    char *sp;
    char *start_p = resp;

    /*
     * I want end_p to actually point one position beyond the last character
     * that is supposed to be in the string.  This makes the rest of the code
     * simpler to read.  The value of len does not seem to account for the
     * NULL character at the end of the string because trimlower is called
     * with things like strlen(resp) as values for len.  So, it should be safe
     * to do this.
     */
    char *end_p = (char *) ((unsigned int)resp + len);

    /*
     * The following code block is a temporary fix that should deal with extra
     * NULLs that are causing a problem when the system, for some unknown
     * reason, prematurely goes into binary mode.  This particular block
     * deals with NULLs at the beginning of the string.  There is already
     * existing code here to deal with other NULLs.  As a beginning condition,
     * start_p must be at the physical beginning of the string that was passed
     * in.
     */
    if (start_p+1 < end_p  &&  *start_p == '\0'  &&  isprint(*(start_p+1)))
	{
	start_p++;
	}

    /*
     * Trim off the leading white space on the string
     */
    while ( (*start_p == '\t' || *start_p == ' ') && start_p != end_p )
	start_p++;

#if 0
    /*
     * This is junk code.  It could result in an end_p that is less than
     * start_p becuase it will decrease sp until j becomes 0.  j, however,
     * is going to be decresed len times.  len is no longer the correct
     * length of the string because we may have increased start_p beyond
     * the beginning of the original string.
     */

    /* Locate the end of the string */
    sp = (char *)  ( (unsigned int)resp + len - 1 );

    for (j=len;j--;j>0)
	{
	if (*sp == '\0' || *sp == '\n' || *sp == '\r')
	    end_p = sp;
	sp--;
	}
#endif

    /*
     * The following code is cleaner and fixes the above problem.  However,
     * I am still not sure whether all of it is necessary, but I will leave
     * it the way it is until I figure out more.
     *
     * As far as I can figure, what the above code is actually doing is
     * finding the FIRST (counting from the beginning up) NULL, new line, or
     * cariage return in resp.  It is doing it very inefficiently.  I will
     * attempt to do it more efficiently.
     */
    sp = start_p;
    while (sp != end_p &&
	   *sp != NULL &&
	   *sp != '\n' && 
	   *sp != '\r')
	sp++;
    
    /*
     * At this point sp should point to a NULL, a new line, a cariage return
     * or whatever end_p is pointing to.  This is where we want end_p to be.
     * And, no matter what this character was before it needs to be a NULL now.
     */
    end_p = sp;
    *end_p = NULL;

    /*
     * Now convert all alpha's to lowercase
     */
    for (sp=start_p; sp<end_p; sp++)
	*sp = tolower (*sp);

    return start_p;    	
    }

/*
 *  Returns:
 *     >0 - 1 based index of the match found between token and
 *		the list array
 *     0  - no match found
 */
int match (char *token, char **list)
    {
    int i=1;
    while ( *list != NULL )
    	{
	if ( strcmp (token, *list) == 0 ) return i;
	list++;
	i++;
    	}

    return 0;
    }

int SetUp_Devices (int *DevFd, int *TtyFd, int BaudRate)
{
    long vdisable;

    if ((*DevFd = get_tpad (DevName)) < 0)
	{
	printf ("? %sLNK All links to the CompuServe network are in use%s",
		PROGRAM_ID_FOR_ERRORS, OUTPUT_LINE_TERMINATOR);
	fflush (stdout);
	return (0);
	}

    if (tcgetattr (*DevFd, &DevSettings) < 0)
	{
	printf ("\nCould not get termios structure");
	fflush (stdout);
	return (0);
	}
    DevTermios = DevSettings;	/* make working copy of structure */
    DevTermios.c_ispeed = BaudRate;
    DevTermios.c_ospeed = BaudRate;

    DevTermios.c_cflag &= ~(CSIZE | PARENB);
    DevTermios.c_cflag |= (CS8);

    DevTermios.c_oflag |= ONLCR;

    DevTermios.c_cc[VMIN] = 0;	/* as much as is ready */
    DevTermios.c_cc[VTIME] = 0;	/* no waiting for input */

    if (tcsetattr (*DevFd, TCSANOW, &DevTermios) < 0)
	{
	printf ("\nCould not set termios structure");
	fflush (stdout);
	return (0);
	}

    if ((*TtyFd = open ("/dev/tty", O_RDWR)) < 0)
	{
	printf ("\nCould not open tty");
	fflush (stdout);
	return (0);
	}

    if (tcgetattr (*TtyFd, &TtySettings) < 0)
	{
	printf ("\nCould not get termios structure");
	fflush (stdout);
	return (0);
	}

    TtyTermios = TtySettings;	/* make working copy of structure */

    /* put tty in raw mode */
    TtyTermios.c_lflag &= ~(ICANON | ECHO | IEXTEN | ISIG);
    TtyTermios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);

    TtyTermios.c_cflag &= ~(CSIZE | PARENB);
    TtyTermios.c_cflag |= (CS8);

    TtyTermios.c_oflag = 0;

    TtyTermios.c_cc[VMIN] = 0;	/* as much as is ready */
    TtyTermios.c_cc[VTIME] = 0;	/* no waiting for input */

    /* ACK -A- HACK! */
    vdisable = _POSIX_VDISABLE;

    TtyTermios.c_cc[VDISCARD] = vdisable;
    TtyTermios.c_cc[VDSUSP] = vdisable;
    TtyTermios.c_cc[VEOF] = vdisable;
    TtyTermios.c_cc[VEOL] = vdisable;
    TtyTermios.c_cc[VEOL2] = vdisable;
    TtyTermios.c_cc[VERASE] = vdisable;
    TtyTermios.c_cc[VINTR] = vdisable;
    TtyTermios.c_cc[VKILL] = vdisable;
    TtyTermios.c_cc[VLNEXT] = vdisable;
    TtyTermios.c_cc[VQUIT] = vdisable;
    TtyTermios.c_cc[VREPRINT] = vdisable;
    TtyTermios.c_cc[VSTART] = vdisable;
    TtyTermios.c_cc[VSTATUS] = vdisable;
    TtyTermios.c_cc[VSTOP] = vdisable;
    TtyTermios.c_cc[VSUSP] = vdisable;
    TtyTermios.c_cc[VWERASE] = vdisable;

    if (tcsetattr (*TtyFd, TCSANOW, &TtyTermios) < 0)
	{
	printf ("\nCould not set termios structure");
	fflush (stdout);
	return (0);
	}

    return (1);
    }

/*
 * Functional Description:
 *	dumpit - displays len bytes of *ptr in hex
 *
 * Formal Parameters:
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

void dumpit (FILE *dev, char *ptr, int len)
    {
    int i;
    
    fprintf (dev, "\n");
    for (i = 0; i < len; i++)
	{
	fprintf (dev, "%02x ", abs (ptr[i]));
	if ( (i != 0) && ((i % 16) == 0) )
	    fprintf (dev, "\n");
	}
    if ( (i-1)%16 != 0)
	fprintf (dev, "\n");
    fflush (dev);
    return;
    }

/*
 * Functional Description:
 *	main
 *
 * Formal Parameters:
 *	argc
 *	argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */
jmp_buf motdinterrupt;

void Display_Text_File (char *filename)
    {
    register int fd, nchars;
    sig_t oldint;
    void sigint ();
    char tbuf[8192];
    char complete_filename[256];

    /*
     * The following code takes the LOCATION_OF_TEXT_FILES constant,
     * which holds the directory name where all the text files used in
     * this program reside, and adds the actual file name to this
     * directory name.  The result should be a correct path and file
     * name stored in the variable complete_filename.  This name is
     * then used as the file to be opened.
     */
    strcpy (complete_filename, LOCATION_OF_TEXT_FILES);
    strcat (complete_filename, filename);

    if ((fd = open (complete_filename, O_RDONLY, 0)) < 0)
        {
        printf ("The file %s could not be opened.\n", complete_filename);
        return;
        }
    oldint = signal (SIGINT, sigint);
    if (setjmp (motdinterrupt) == 0)
        while ((nchars = read (fd, tbuf, sizeof (tbuf))) > 0)
            (void) write (fileno (stdout), tbuf, nchars);

    fflush (stdout);

    (void) signal (SIGINT, oldint);
    (void) close (fd);
    }

void sigint ()
    {
    longjmp (motdinterrupt, 1);
    }

/*
 * Functional Description:
 *      Initial_Connect
 * At connect, print out some debug stuff (like where I think you're
 * coming from) and map the inbound device name (ttyxx) to a pad
 * device name (tpadxx).
 *
 *
 * Formal Parameters:
 *
 * Implicit Parameters:
 *
 * Return Value:
 *
 * Side Effects:
 *
 */
int Initial_Connect (char *DevName, char *oDevName)
    {
    showoriginip;
    printf ("DevName  = [%s]\noDevName = [%s]\n", DevName, oDevName);
    /*
     * We want to print to the user, give the lines a moment to settle down.
     */
    fflush (stdout);
    sleep (1);
    return 0;
    }

/*
 * Functional Description:
 *      main
 *
 * Formal Parameters:
 *      argc
 *      argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

/*
 * Do interactive session with a user, prior to granting access to the
 * information service.
 */

int Perform_Dialogs ()
    {
    if (!Get_First_Prompt ())
        {
        sleep (1);
        exit (0);
        }

    puts ("\nPLEASE NOTE: You can skip this introductory information by "
          "entering CISAGREE\nat the Host Name prompt in the future.");
    /*
     * We don't display the baud rate prompt anymore.
     */
#if 0
    Display_Text_File (ExplainBitRate);
#endif

    return 0;
    }

/*
 * Functional Description:
 *      main
 *
 * Formal Parameters:
 *      argc
 *      argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

/* Get a password for beta test purposes */

int Get_Password ()
{
    char resp[128];
    char *sp;
    int cnt = 0;
    int TtyFd;
    struct termios TtyTermios;

    if ((TtyFd = open ("/dev/tty", O_RDWR)) < 0)
        {
        printf ("\nCould not open tty");
        fflush (stdout);
        exit (1);
        }

    resp[0] = '\0';
    sp = trimlower(resp, sizeof(resp));

    while (strncmp (sp, "not a chance!", 13) && cnt < 3)
        {
        printf ("Please enter test password: (required only during beta test) ");

        if (tcgetattr (TtyFd, &TtySettings) < 0)
            {
            printf ("\nCould not start to turn off echo");
            exit (1);
            }

        TtyTermios = TtySettings;       /* make working copy of structure */
        TtyTermios.c_lflag &= ~ECHO;

        if (tcsetattr (TtyFd, TCSANOW, &TtyTermios) < 0)
            {
            printf ("\nCould not turn off echo");
            fflush (stdout);
            exit (1);
            }

        fflush (stdout);

        resp[0] = '\0';
        if ( NULL == fgets (resp, 128, stdin))
            {
            exit (-1);
            }
        sp = trimlower(resp, sizeof(resp));

        if (tcgetattr (TtyFd, &TtySettings) < 0)
            {
            printf ("\nCould not start to turn on echo");
            fflush (stdout);
            exit (1);
            }

        TtyTermios = TtySettings;       /* make working copy of structure */
        TtyTermios.c_lflag |= ECHO;

        if (tcsetattr (TtyFd, TCSANOW, &TtyTermios) < 0)
            {
            printf ("\nCould not turn on echo");
            fflush (stdout);
            exit (1);
            }
        printf ("\n");
        fflush (stdout);
        cnt++;
        }

    close (TtyFd);

    if (strncmp (resp, "not a chance!", 13) == 0)
      {
        Display_Text_File (GetPasswordGood);
        return (1);
      }

    Display_Text_File (GetPasswordBad);

    return (0);
}

/*
 * Functional Description:
 *      main
 *
 * Formal Parameters:
 *      argc
 *      argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

int Get_First_Prompt ()
{
    char resp[128];
    char *sp;

    /*
     * This prompt is for agreement to abide by whatever policies and
     * rules the orginating domain may impose on the user.
     */

    Display_Text_File (GetFirstPrompt1);

  I_Hate_Gotos:
    Display_Text_File (GetFirstPrompt2);

    resp[0] = '\0';
    printf ("Enter choice (LOGON, HELP, OFF): ");
    if ( NULL == fgets (resp, 128, stdin)) exit (-1);

    sp = trimlower(resp, sizeof(resp));

    while ( (0 != strcmp(sp, "off")) &&
           (0 != strcmp(sp, "bye")) &&
           (0 != strcmp(sp, "exit")) &&
           (0 != strcmp(sp, "help")) &&
           (0 != strcmp(sp, "logon")) &&
           (0 != strcmp(sp, "?")))
        {
        printf ("You must enter either OFF, LOGON, or HELP.\n");
        Display_Text_File (GetFirstPrompt2);

        printf ("Enter choice (LOGON, HELP, OFF): ");

        fflush (stdout);
        if ( NULL == fgets (resp, 128, stdin)) exit(-1);

#if 0
        printf ("Recieved: ");
        dumpit (stdout, resp, strlen(resp)+2);
#endif
        
        sp = trimlower(resp, sizeof(resp));
        }

    if (0 == strcmp(sp, "logon"))
        return (1);

    if (strcmp(sp, "help") && *sp != '?')
        {
        printf ("We are terminating the connection at this time as you requested.\n");
        fflush (stdout);
        return (0);
        }

    if (!Get_Help ())
        return (0);
    goto I_Hate_Gotos;

}

int Get_Agreement_Text ()
    {
    char resp[128];
    char *sp;

    /*
     * This prompt is for agreement to abide by whatever policies and
     * rules the orginating domain may impose on the user.
     */

    Display_Text_File (GetAgreeText1);
    printf ("Press the Enter or Return key for more: ");
    if (NULL == fgets(resp,128,stdin))
        exit (-1);
    sp = trimlower(resp,128);

    while ((0 != strcmp(sp, "off")) &&
           (0 != strcmp(sp, "bye")) &&
           (0 != strcmp(sp, "exit")) &&
           (0 != strcmp(sp, "help")) &&
           (0 != strcmp(sp, "?")) &&
           (*sp != NULL))
        {
        printf ("Enter OFF to terminate the connection or press the Enter or Return key");
        printf ("to continue.\n\n");
        printf ("Press the Enter or Return key for more: ");
        if ( NULL == fgets( resp, 128, stdin)) exit (-1);
        sp = trimlower( resp, 128);
        }

    if (*sp !='\0' && strcmp(sp,"help") && *sp != '?')
        {
        printf ("We are terminating the connection at this time as you requested.\n");
        fflush (stdout);
        return (0);
        }

    Display_Text_File (GetAgreeText2);
    printf ("Press the Enter or Return key for more: ");
    if ( NULL == fgets( resp, 128, stdin)) exit (-1);
    sp = trimlower( resp, 128);

    while ( (0 != strcmp(sp, "off")) &&
           (0 != strcmp(sp, "bye")) &&
           (0 != strcmp(sp, "exit")) &&
           (0 != strcmp(sp, "help")) &&
           (0 != strcmp(sp, "?")) &&
           (*sp != '\0'))
        {
        puts ("Enter OFF to terminate the connection or press the Enter or"
              " Return key\nto continue.\n");
        printf ("Press the Enter or Return key for more: ");
        if ( NULL == fgets( resp, 128, stdin)) exit (-1);
        sp = trimlower( resp, 128);
        }

    if (*sp !='\0' && strcmp(sp,"help") && *sp != '?')
      {
        printf ("We are terminating the connection at this time as you requested.\n");
        fflush (stdout);
        return (0);
      }

    Display_Text_File (GetAgreeText3);
    return (1);
  }

/*
 * Functional Description:
 *      main
 *
 * Formal Parameters:
 *      argc
 *      argv
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *
 *
 * Side Effects:
 *
 */

int Get_Agreement ()
{
    char resp[128];
    char *sp;
    int ic = 0;

    resp[0] = '\0';
    sp = resp;

    if (!Get_Agreement_Text()) return(0);

    printf ("Enter choice (AGREE or OFF): ");
    if ( NULL == fgets( resp, 128, stdin))
        exit (-1);
    sp = trimlower( resp, 128);

    while (strcmp(sp,"agree") != 0 &&
           strcmp(sp,"bye")   != 0 &&
           strcmp(sp,"exit")  != 0 &&
           strcmp(sp,"off")   != 0 &&
           ic < 3)
        {
        puts ("You must enter AGREE to indicate that you have reviewed"
              " these terms, or\nelse enter OFF to terminate the"
              " connection.\n");

        if (!Get_Agreement_Text())
            return(0);
        printf ("Enter choice (AGREE or OFF): ");
        fflush (stdout);
        if ( NULL == fgets (resp, 128, stdin))
            exit (-1);
        sp = trimlower( resp, 128);
        ic++;
        }

    if (strcmp (sp, "agree") != 0)
        {
        printf ("We are terminating the connection at this time as you requested.\n");
        fflush (stdout);
        return (0);
        }

    printf ("PLEASE NOTE:  You can skip the review of these precautions by entering\n");
    printf ("CISAGREE at the Host Name prompt in the future.\n");

    /* This is the BIT Rate Explanation */
    Display_Text_File (ExplainBitRate);

    return (1);
    }

/*
 * Functional Description:
 *      Put up a "menu" of informational topics that the user can request
 *      to have mailed to their internet mail network address. The menu
 *      support logic is *very* rudimentary.
 *
 * Formal Parameters:
 *      None
 *
 * Implicit Parameters:
 *      
 *
 * Return Value:
 *      1 - user wants to go back to the membership prompt
 *      0 - disconnect user from this session
 *
 * Side Effects:
 *      records the origin ip address and the target mail request specified
 *      in the syslog facility.
 *      makes a system call to "/usr/bin/mail"
 */
int Get_Help ()
    {
    char resp[128];
    char file[128];
    char *sp;
    int mail_selection[8] =
    {0, 0, 0, 0, 0, 0, 0, 0};
    int mail_selection_size = (sizeof(mail_selection)/sizeof(int));
    int i;

    /*
     * Help introduction...
     */
    Display_Text_File (GetHelpText1);
    Display_Help_Menu ();

    resp[0] = 'x';  /* Force the next loop to execute the first time */
    sp = trimlower(resp, sizeof(resp));
    i=0;

    /*
     * Process the menu request.
     */
    while (*sp != '1' && strcmp(sp, "off") &&
           strcmp(sp, "bye") && strcmp(sp, "exit") &&
           *sp != '2' && *sp != '3' &&
           *sp != '4' && *sp != '5' &&
           *sp != '6' && *sp != '\0')
        {
        if (i > 0)
        printf ("\nYou must enter a blank line, number(s) between 1 and 6,"
                " or OFF.\n");

        i++;

/*      Display_Text_File (GetHelpText2); */
        printf (GetHelpPrompt1);
        fflush (stdout);
        if ( NULL == fgets (resp, 128, stdin)) exit (-1);
        sp = trimlower(resp, sizeof(resp));
        }

    if (*sp == '\0') 
        {
        printf ("You have not selected any articles.\n");
        return 1;
        }

    if (0 == strcmp(sp, "off") ||
        0 == strcmp(sp,"exit") ||
        0 == strcmp(sp, "bye") )
        {
        puts ("We are terminating the connection at this time as you "
              "requested.");
        fflush (stdout);
        return (0);
        }

    /* Set up the list of requested entries */
    for (i = 0; i < strlen (sp); i++)
        {

        if (!isdigit(sp[i]))
            {
            sp[i] = ' ';
            }

        switch (sp[i])
            {
            case ',':
            case ' ':
                break;
            case '1':
                mail_selection[1] = 1;
                break;
            case '2':
                mail_selection[2] = 1;
                break;
            case '3':
                mail_selection[3] = 1;
                break;
            case '4':
                mail_selection[4] = 1;
                break;
            case '5':
                mail_selection[5] = 1;
                break;
            case '6':
                mail_selection[6] = 1;
                break;
            default:
                break;
            }
        }

    /*
     * Tell the user about our plan to internet mail the information
     * to them, and then get their address
     */
    Display_Text_File (GetHelpText3);

    resp[0] = '\0';

    printf ("Enter choice (blank line, Electronic Mail Address or Off): ");
    fflush (stdout);
    if ( NULL == fgets (resp, 128, stdin)) exit (-1);
    sp = trimlower(resp, sizeof(resp));

    if (0 == strcmp(sp, "off") ||
        0 == strcmp(sp, "exit") ||
        0 == strcmp(sp, "bye"))
        {
        printf ("We are terminating the connection at this time as you "
                "requested.\n");
        fflush (stdout);
        return (0);
        }

    if (*sp == '\0')
        {
        printf ("\nNo articles will be sent to you.\n\n");
        fflush (stdout);
        return(1);
        }

    /*
     * This is where we would check for an address format if there was a
     * reasonablness test available...
     */

    for (i = 1; i < mail_selection_size; i++)
        {
        if (mail_selection[i])
            {
            strcpy (file, LOCATION_OF_TEXT_FILES);
	    strcat (file, "mail_selection.0");
	    file[strlen (file) - 1] = '0' + i;
		{
		char buf[128];

		syslog (LOG_INFO, "ip: %s  mailaddr: %s", IP_Address,
			sp);

		sprintf (buf, "cat %s 2>/dev/null | /usr/bin/mail "
			 "-s \"CIS Information\" %s >/dev/null 2>&1",
			 file, sp);

		system (buf);
		}
	    }
	}

    Display_Text_File (GetHelpText4);
    return (1);
}

void timedout ()
    {
    printf ("\r\n\nCannot take more than 2 minutes to Log-in\r\n\n");
    fflush (stdout);
    exit (EXIT_FAILURE);
    }

/*
 * Functional Description:
 *	Simulate the CompuServe Network Host Name prompt. Doing so
 *	allows script developers to write scripts now that will still
 *	work if BSG access is permitted in the future.
 *
 * Formal Parameters:
 *	none
 *
 * Implicit Parameters:
 *
 *
 * Return Value:
 *       HOSTNAME_BAD - selected service is not one of:
 *		CIS, CPS, or CISAGREE
 *       HOSTNAME_CIS - user selected CIS or CPS service
 *	 HOSTNAME_AGREE - user entered CISAGREE
 *
 * Side Effects:
 *       performs an I/O request to the user
 */

int Get_Service ()
    {
    char resp[128];      /* A response from the user. */
    char *sp;            /* I suppose this stands for string pointer. */
    int service_index;   /* Index to the HostNameList array. */

    /*
     * Set up the signal handler function, timeout, for the signal SIGALRM,
     * which will be generated if the alarm that we will set goes off.
     */
    (void) signal (SIGALRM, timedout);

    /*
     * The following while loop forces the user to enter correct input at the
     * "Host Name:" prompt that it displays.
     */
    while (1)
	{

	/*
	 * We are going to display a new "Host Name:" prompt so we must reset
	 * the timeout alarm.
	 */
	alarm (120);
	printf (HostNamePrompt);
        if (NULL == fgets(resp,128,stdin))  /* Read in the users response. */
	    {

	    /*
	     * This will happen if they press a Control-D?
	     */
	    exit (EXIT_SUCCESS);
	    }

	/*
	 * We recieved some input so we better turn off the alarm now.  It will
	 * get turned back on at the top of the loop if the top of the loop is
	 * reached again.
	 */
	alarm (0);

        sp = trimlower(resp, sizeof(resp));  /* resp is no longer valid */
	service_index = match(sp, HostNameList);
	if (service_index)
	    {
	    /*
	     * A valid host name (service) was entered.
	     */
	    signal (SIGALRM, SIG_DFL);  /* reset handler before return */
	    return (service_index);
	    }
	else if (!strcmp(sp,"off") || !strcmp(sp,"exit") || !strcmp(sp,"bye"))
	    {
	    puts ("We are terminating the connection at this time as you "
		  "requested.");
	    fflush (stdout);
	    signal (SIGALRM, SIG_DFL);  /* reset handler before return */
	    return (HOSTNAME_BAD);
	    }

	/*
	 * Nothing useful was entered, so display a help message if the
	 * response is not null.
	 */
	if (strcmp(sp,"") != 0)
	    {
	    Display_Text_File ("inetcis.text.host_name_help");
	    }
	}  /* end while */
    }

void Display_Help_Menu ()

/*
 * Functional Description:
 *
 *     Displays the help menu, including the date and time of last
 *     of each of the selections.
 *
 * Formal Parameters:
 *
 *     none
 *
 * Implicit Parameters:
 *
 *     none
 *
 * Return Value:
 *
 *     none
 *
 * Side Effects:
 *
 *     none
 */

    {
    const char help_menu_pretext[] = "\nTo obtain additional information,"
    " select the article(s) of interest from the\nfollowing list and they"
    " will be sent to you via electronic mail.\n";
    const char *help_menu_selection[] = 
	{
	" 1 CompuServe Overview",
	" 2 CompuServe Information Service Overview",
	" 3 How To Obtain a CompuServe Information Service Membership",
	" 4 CompuServe Information Service Rates",
	" 5 Exchanging E-Mail between Internet and CompuServe",
	" 6 Internet Press Release"
	};
    const int number_of_selections = 6;
    const char help_menu_posttext[] = "\nTo select items from the menu, enter "
    "one or more of the corresponding numbers\nseparated by commas and "
    "followed by the Enter or Return key.  To leave this\nmenu without "
    "selecting any articles, simply press the Enter or Return key.\nTo "
    "terminate your connection, enter OFF followed by the Enter or Return "
    "key.\n";
    int i;
    struct stat help_file_info;
    char help_file_name[256];
    struct tm *modification_time_ptr;
    char time_output_string[256];

    puts (help_menu_pretext);
    fflush (stdout);
    strcpy (help_file_name, LOCATION_OF_TEXT_FILES);
    strcat (help_file_name, "mail_selection.0");
    for (i=1; i<=number_of_selections; i++)
	{

	/*
	 * Convert i to a char and add it to the name of the help file.
	 */
	help_file_name[strlen(help_file_name) - 1] = (char) i+48;

	/*
	 * This is a POSIX system call that gets information about a file and
	 * puts it in the structure help_file_info.  It is defined in
	 * sys/stat.h.
	 */
	if ( stat(help_file_name,&help_file_info) == 0 /* 0 is NO ERROR */)
	    {
	    /* 
	     * Put the time in struct tm format instead of time_t format so the
	     * individual pieces can be accessed.
	     */
	    modification_time_ptr = localtime(&help_file_info.st_mtime);

	    /*
	     * Convert the modification time to the format that is needed for
	     * displaying it.
	     */
	    strftime (time_output_string, 10, "%d-%b-%y",
		      modification_time_ptr);
	    }
	else
	    {
	    strcpy (time_output_string, "not available");
	    }

	/*
	 * Display the whole line.
	 */
	printf ("%s (%s)\n", help_menu_selection[i-1], time_output_string);
	fflush (stdout);
	}
    puts (help_menu_posttext);
    fflush (stdout);
    }




void Read_Asterisk_Prompt ()
    {
    int character_count;
    int stripped_character_count;
    char stripped_buffer[BUFSIZE];
    char buffer[BUFSIZE];

    /*
     * The following code sets up an alarm to interrupt the read that comes
     * after it.
     */

    if (signal (SIGALRM, Alarm_Signal_Handler) == SIG_ERR)
	{
	syslog (LOG_ERR, "before line %d signal caused: %m", __LINE__);
	fprintf (stderr, "? %sSIG Failed to assign signal handler%s",
		 PROGRAM_ID_FOR_ERRORS, OUTPUT_LINE_TERMINATOR);
	exit (EXIT_FAILURE);
	}
    if (setjmp(asterisk_read_timeout) != 0)
	{
	fprintf (stderr, "? %sATM Timeout waiting for response from node%s",
		 PROGRAM_ID_FOR_ERRORS, OUTPUT_LINE_TERMINATOR);
	syslog (LOG_ERR, "timed out waiting for asterisk prompt in"
		" Read_Asterisk_Prompt");
	exit (EXIT_FAILURE);
	}

    alarm (30);

    /*
     * The following reads from the X.25 device repeatedly.  The device is in
     * non-blocking mode, so when the read is done it will immediately return
     * if there is nothing there, so it must loop until it gets something.
     * The first piece of data should be the echo of the xset command done
     * above, but we don't care about that.  All we care about is getting the
     * asterisk prompt back so we can do the following write.  That is what
     * the loop will wait for.  If it doesn't read it in a certain amount of
     * time is should interrupt itself.
     */
    
    stripped_character_count = 1;  /* must be non-zero for first strncmp */
    strcpy (stripped_buffer, "");
    while (strncmp("*",stripped_buffer,stripped_character_count) != 0)
	{
	character_count = read (DevFd, buffer, BUFSIZE);
	if (character_count < 0)
	    {
	    if (errno != EWOULDBLOCK)
		{
		syslog (LOG_ERR, "before line %s read caused: %m", __LINE__);
		fprintf (stderr, "? %sREN Error reading from network%s",
			 PROGRAM_ID_FOR_ERRORS, OUTPUT_LINE_TERMINATOR);
		exit (EXIT_FAILURE);
		}
	    }
	else
	    {
	    Strip_Leading_CRs_And_LFs (buffer, stripped_buffer,
				       character_count,
				       &stripped_character_count);
	    }
	}  /* while */

    alarm (0);
    if (signal(SIGALRM,SIG_DFL) == SIG_ERR)
	{
	syslog (LOG_ERR, "before line %d signal caused: %m", __LINE__);
	fprintf (stderr, "? %sSG2 Failed to assign signal handler%s",
		 PROGRAM_ID_FOR_ERRORS, OUTPUT_LINE_TERMINATOR);
	exit (EXIT_FAILURE);
	}
    }  /* end of Read_Asterisk_Prompt */


void Alarm_Signal_Handler (int signo)
    {
    longjmp (asterisk_read_timeout, 1);
    }



void Write_All_Characters (FILE *f, char *message, char *characters,
			   int how_many)
    {
    int i;

    fprintf (f, "%s", message);

    for (i=0; i < how_many; i++)
	{
	if (isprint(characters[i]))
	    {
	    fputc ((int) characters[i], f);
	    }
	else if (characters[i] == '\r')
	    {
	    fprintf (f, "[CR]");
	    }
	else if (characters[i] == '\n')
	    {
	    fprintf (f, "[LF]");
	    }
	else if (characters[i] == '\0')
	    {
	    fprintf (f, "[NULL]");
	    }
	else
	    {
	    fprintf (f, "[%d]", characters[i]);
	    }
	}
    fputc ('\n', f);
    }



void exit_handler ()
    {
    pid_t pid_of_inetcis_telnetd;
    int return_value;
    const int KILL_ERROR = -1;

    /*
     * The getppid() function returns the process ID (pid) of this process'
     * parent process.  It is always successful.  No error checking is
     * required.  inetcis.telnetd should be the parent of this process.
     */
    pid_of_inetcis_telnetd = getppid();

    /*
     * Because a pty/tty communication facility is being used between
     * inetcis.telnetd and inetcis.login, inetcis.login needs to send a signal
     * to inetcis.telnetd to indicate that the tty is being closed.
     * inetcis.telnetd has no other way of finding out that the tty
     * connected to its pty is being closed.  Since there is no reason for
     * inetcis.telnetd to continue running, SIGTERM is a fine signal to send.
     *
     * Well, it seems that we'll have to get more drastic.  Since SIGTERM is
     * not working in all cases, I'll send a SIGKILL.  HA!
     */
    return_value = kill (pid_of_inetcis_telnetd, SIGKILL);
    if (return_value == KILL_ERROR)
	{
	switch (errno)
	    {
	  case ESRCH:
	    syslog (LOG_INFO, "Kill was not able to send signal to "
		    "parent because parent has already "
		    "terminated.");
	    break;
	  case EPERM:
	    syslog (LOG_CRIT, "inetcis.telnetd is not running as the same "
		    "user as inetcis.login so inetics.login cannot signal "
		    "inetcis.telnetd to terminate.  inetcis.telnetd's will "
		    "will accumulate until no more can be started.");
	    break;
	  default:
	    syslog (LOG_ERR, "Kill inetcis.telnetd returned error: %m");
	    break;
	    }
	}
    else
	{
	syslog (LOG_INFO, "Terminate siginal sent to parent.");
	}

    syslog (LOG_DEBUG, "exiting");
    closelog ();
    }


void Get_Baud_Rate_from_User (int *baud, int *baud_rate, int *max_char)
    {
    char response[256];
    char *response_ptr;
    int number_of_tries = 0;

    *baud_rate = B0;  /* B0 from termios.h */
    do 
	{
	printf ("Enter choice (300, 2400, 9600, 14400, OFF): ");
	gets (response);
	response_ptr = trimlower(response, sizeof(response));

	if ( ! strcmp (response_ptr, "300")  ||
	     ! strcmp (response_ptr, "450")  ||
	     ! strcmp (response_ptr, "1200") ||
	     ! strcmp (response_ptr, "1800") )
	    {
	    *baud_rate = B300;  /* B300 from termios.h */
	    }
	else if ( ! strcmp (response_ptr, "2400") ||
		  ! strcmp (response_ptr, "4800") )
	    {
	    *baud_rate = B2400;  /* B2400 from termios.h */
	    }
	else if ( ! strcmp (response_ptr, "9600")     ||
		  ! strcmp (response_ptr, "14400")    ||
		  ! strcmp (response_ptr, "19200")    ||
		  ! strcmp (response_ptr, "28800")    ||
		  ! strcmp (response_ptr, "38400")    ||
		  ! strcmp (response_ptr, "57600")    ||
		  ! strcmp (response_ptr, "115200")   ||
		  ! strcmp (response_ptr, "quick")    ||
		  ! strcmp (response_ptr, "infinity") )
	    {
	    *baud_rate = B9600;  /* B9600 from termios.h */
	    }
	else if ( ! strcmp (response_ptr, "off")  ||
		  ! strcmp (response_ptr, "exit") ||
		  ! strcmp (response_ptr, "bye") )
	    {
	    puts ("We are terminating the connection at "
				    "this time as you requested.");
	    sleep (1);  /* not completely sure why this is here */
	    exit (EXIT_SUCCESS);
	    }
	else  /* bad input */
	    {
	    puts ("You must enter either a valid baud rate "
				    "or OFF.");
	    Display_Text_File (ExplainBitRate);

	    if (++number_of_tries > 2)
		{
		puts ("Valid baud rate not entered.  "
		      "Terminating connection.");
		exit (EXIT_FAILURE);
		}
	    }
	    
	} while (*baud_rate == B0);  /* B0 from termios.h */

    /*
     * Since we are no longer doing throttling, set *max_char to a large
     * number.  This, in the past, was set to *baud_rate/10.
     */
    
    *max_char = INT_MAX;

    /* 
     * *baud is not being used now but I'm leaving it in for now just
     * in case.
     */
    *baud = (int) response_ptr[0];

    syslog (LOG_INFO, "baud rate %s reported as %d", response_ptr, *baud_rate);
    } /* Get_Baud_Rate_From_User */



void terminate_handler (int signal_number)
    {
    syslog (LOG_INFO, "%s %d - info - caught SIGTERM signal", __FILE__,
	    __LINE__);
    exit (EXIT_SUCCESS);
    }
