/*
 * Copyright (c) 1995 by CompuServe Incorporated, Columbus, Ohio
 * 
 * Facility:  ?
 * 
 * Abstract:
 * 
 * ?
 * Environment:  ?
 * 
 * Author:  ?
 * 
 * Revision History:
 * 
 * 1.0  ?, ? - Initial revision
 */

/*
 * External Interfaces
 */
#include <syslog.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <comdef.h>		/* CompuServe definitions */
#include <log.h>
#include <log_pipe.h>
#include <poserr.h>

/*
 * Module-wide definitions
 */
#define MAX_STRING_LENGTH  ((size_t) 255)
#define STRING_BUFFER_SIZE ((size_t) MAX_STRING_LENGTH + 1)

static FILE    *local_log_file_ptr;
static char     compuserve_program_id[4];
static FILE    *user_ptr;
static bool     user_log_is_active = FALSE;
static char     user_log_line_terminator[3];
static bool     local_file_log_is_active = FALSE;
static bool     syslog_log_is_active = FALSE;
static int      user_fd_log_fd;
static bool     user_fd_log_is_active = FALSE;
static int      inhouse_stderr_log_is_active = FALSE;
static char     log_level_description[11][12] =
{
  "EMERGENCY", "ALERT", "CRITICAL", "ERROR", "WARNING", "NOTICE",
  "INFORMATION", "DEBUG0", "DEBUG1", "DEBUG2", "DEBUG3"
};

/*
 * Global data
 */

/*
 * Table of contents
 */
int             LogLevel_To_Syslog_Priority (LogLevel level);
int             Write_All (int fd, char *buffer, int n);


void 
Start_Local_File_Log (char *directory, char *file_name_prefix)
{
  time_t          encoded_time;
  struct tm      *struct_time_ptr;
  char            formatted_time[10];
  char           *local_log_file_name = NULL;
  char           *directory_with_slash = NULL;
  char           *directory_name_buffer = NULL;

  encoded_time = time (NULL);
  struct_time_ptr = localtime (&encoded_time);
  (void) strftime (formatted_time, 10, "%H:%M:%S", struct_time_ptr);

  if (directory[strlen (directory) - 1] == '/') {	/* check dir name */
    directory_with_slash = directory;
  } else {
    directory_name_buffer = (char *) malloc (sizeof (char) *
					     (strlen (directory) + 1));
    /*
     * This step is necessary.  It allows directory_name_buffer to be freed,
     * as opposed to freeing directory_with_slash.  When it comes time to
     * free this space, it will not be known if directory_with_slash points
     * to malloced memory (directory_name_buffer) or static memory
     * (directory).
     */
    directory_with_slash = directory_name_buffer;
    strcpy (directory_with_slash, directory);
    strcat (directory_with_slash, "/");
  }

  local_log_file_name = (char *) malloc (sizeof (char) * (
					     strlen (directory_with_slash) +
						 strlen (file_name_prefix) +
					     strlen (formatted_time) + 20));

  if (local_log_file_name == NULL) {	/* check for malloc error */
    Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			"local log file name", "malloc", "MEM",
			"Could not allocate memory");
    return;
  }
  sprintf (local_log_file_name, "%s%s.%d.log", directory_with_slash,
	   file_name_prefix, (int) getpid ());

  if (directory_name_buffer != NULL) {	/* free mem if it was allocated */
    free (directory_name_buffer);
  }
  local_log_file_ptr = fopen (local_log_file_name, "w");
  free (local_log_file_name);

  if (local_log_file_ptr == NULL) {	/* check for fopen error */
    Log_Hidden_System_Error (LOG_LEVEL_ERR, __FILE__, __LINE__,
			     errno, "local log file", "fopen",
			     "Could not open log file");
    return;
  }
  local_file_log_is_active = TRUE;

  Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			"local log file is opened");
}

void 
Start_Pipe_Log (char *pipe_command)
{
  Open_Debug_Pipe (pipe_command);
  Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			"pipe log is opened");
}

void 
Start_User_Log (FILE * user_file_ptr, char *three_letter_program_id,
		char *line_terminator)
{
  if (user_file_ptr == NULL) {
    return;
  }
  strncpy (user_log_line_terminator, line_terminator, 2);
  user_log_line_terminator[2] = EOS;
  user_ptr = user_file_ptr;
  strncpy (compuserve_program_id, three_letter_program_id, 3);
  compuserve_program_id[3] = EOS;
  Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			"user messages are activated");

  user_log_is_active = TRUE;
}

void 
Start_User_Fd_Log (int fd, char *three_letter_program_id,
		   char *line_terminator)
{
  strncpy (user_log_line_terminator, line_terminator, 2);
  user_log_line_terminator[2] = EOS;
  user_fd_log_fd = fd;
  strncpy (compuserve_program_id, three_letter_program_id, 3);
  compuserve_program_id[3] = EOS;
  Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			"user file descriptor messages are activated");

  user_fd_log_is_active = TRUE;
}

void
Start_Inhouse_Stderr_Log (char *cs_program_id)
{
  strncpy (compuserve_program_id, cs_program_id, 3);
  compuserve_program_id[3] = EOS;
  inhouse_stderr_log_is_active = TRUE;
}

void 
Stop_User_Fd_Log ()
{
  user_fd_log_is_active = FALSE;
}

void 
Start_Syslog_Log (char *program_name)
{
  if (syslog_log_is_active) {
    return;
  }
  openlog (program_name, LOG_CONS | LOG_NDELAY | LOG_PID, LOG_LOCAL7);
  syslog_log_is_active = TRUE;
  Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			"syslog loging is activated");
}

void 
Write_Va_Message_To_Log (LogLevel level, char *source_file_name,
			 int source_file_line_number,
			 char *message,...)
{
  va_list         argument_list;
  char            formatted_message[STRING_BUFFER_SIZE];
  int             status;

  va_start (argument_list, message);
  status = vsnprintf (formatted_message, STRING_BUFFER_SIZE, message,
		      argument_list);
  if (status >= STRING_BUFFER_SIZE) {
    Write_Message_To_Log (LOG_LEVEL_INFO, __FILE__, __LINE__,
			  "the following message is truncated");
  }
  va_end (argument_list);

  if (syslog_log_is_active && level < LOG_LEVEL_DEBUG0) {
    syslog (LogLevel_To_Syslog_Priority (level),
	    "%s: %s, file: %s, line: %d", log_level_description[level],
	    formatted_message,
	    source_file_name, source_file_line_number);
  }
  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr, "%s: %s, file: %s, line: %d\n",
	     log_level_description[level], formatted_message,
	     source_file_name, source_file_line_number);
    fflush (local_log_file_ptr);
  }
  if (Debug_Pipe_Is_Open ()) {
    Write_To_Debug_Pipe (formatted_message);
  }
}


void 
Write_Message_To_Log (LogLevel level, char *source_file_name,
		      int source_file_line_number,
		      char *message)
{
  if (syslog_log_is_active && level < LOG_LEVEL_DEBUG0) {
    syslog (LogLevel_To_Syslog_Priority (level),
	    "%s: %s, file: %s, line: %d", log_level_description[level],
	    message, source_file_name, source_file_line_number);
  }
  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr, "%s: %s, file: %s, line: %d\n",
	     log_level_description[level], message, source_file_name,
	     source_file_line_number);
    fflush (local_log_file_ptr);
  }
  if (Debug_Pipe_Is_Open ()) {
    Write_To_Debug_Pipe (message);
  }
}

void 
Write_Error_To_Log (LogLevel level,
		    char *source_file_name,
		    int source_file_line_number,
		    char *object_description,
		    char *operation_description,
		    char *compuserve_three_letter_error_id,
		    char *message)
{
  if (syslog_log_is_active && level < LOG_LEVEL_DEBUG0) {
    syslog (LogLevel_To_Syslog_Priority (level),
	    "%s: %s, object: %s,"
	    " operation: %s, file: %s, line: %d",
	    log_level_description[level],
	    message, object_description,
	    operation_description, source_file_name,
	    source_file_line_number);
  }
  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr,
	     "%s: %s, object: %s,"
	     " operation: %s, file: %s, line: %d\n",
	     log_level_description[level],
	     message,
	     object_description,
	     operation_description,
	     source_file_name,
	     source_file_line_number);
    fflush (local_log_file_ptr);
  }
  if (user_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    fprintf (user_ptr, "%s %s%s - %s%s", compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     message, user_log_line_terminator);
  }
  if (inhouse_stderr_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    fprintf (stderr, "%s %s%s - %s for %s with %s\n\tfile: %s\n\tline: %d\n",
	     compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     message, object_description,
	     operation_description, source_file_name, source_file_line_number);
  }
  if (user_fd_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;
    char            message_line[1024];

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    sprintf (message_line, "%s %s%s - %s%s",
	     compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     message, user_log_line_terminator);
    Write_All (user_fd_log_fd, message_line, strlen (message_line));
  }
}				/* Write_Error_To_Log() */

void 
Write_System_Error_To_Log (LogLevel level,
			   char *source_file_name,
			   int source_file_line_number,
			   int system_error_number,
			   char *object_description,
			   char *operation_description,
			   char *compuserve_three_letter_error_id,
			   char *message)
{
  if (syslog_log_is_active && level < LOG_LEVEL_DEBUG0) {
    syslog (LogLevel_To_Syslog_Priority (level),
	    "%s: %s, system message: %s, object: %s,"
	    " operation: %s, file: %s, line: %d",
	    log_level_description[level],
	    message, strerror (errno), object_description,
	    operation_description, source_file_name,
	    source_file_line_number);
  }
  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr, "%s: %s\n\tsystem"
	     " message:\t%s\n\tobject:\t%s\n\t"
	     "operation:\t%s\n\tfile:\t%s\n\tline\t%d\n",
	     log_level_description[level],
	     message, strerror (errno), object_description,
	     operation_description, source_file_name,
	     source_file_line_number);
    fflush (local_log_file_ptr);
  }
  if (user_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    fprintf (user_ptr, "%s %s%s - %s%s", compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     message, user_log_line_terminator);
  }
  if (inhouse_stderr_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    fprintf (stderr, "%s %s%s - %s for %s with %s\n\tfile: %s\n\tline: %d\n",
	     compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     strerror(system_error_number), object_description,
	     operation_description, source_file_name, source_file_line_number);
  }
  if (user_fd_log_is_active && level < LOG_LEVEL_NOTICE) {
    char           *compuserve_priority_identifier;
    char            message_line[1024];

    switch (level) {
    case LOG_LEVEL_WARNING:
      compuserve_priority_identifier = "%";
      break;
    case LOG_LEVEL_ERR:
      compuserve_priority_identifier = "?";
      break;
    default:
      compuserve_priority_identifier = "??";
    }
    sprintf (message_line, "%s %s%s - %s%s",
	     compuserve_priority_identifier,
	     compuserve_program_id, compuserve_three_letter_error_id,
	     message, user_log_line_terminator);
    Write_All (user_fd_log_fd, message_line, strlen (message_line));
  }
}				/* Write_System_Error_To_Log() */


/* Log_Hidden_System_Error should log the message to all destinations
   except directly to the user. */
void 
Log_Hidden_System_Error (LogLevel level,
			 char *source_file_name,
			 int source_file_line_number,
			 int system_error_number,
			 char *object_description,
			 char *operation_description,
			 char *message)
{
  if (syslog_log_is_active && level < LOG_LEVEL_DEBUG0) {
    syslog (LogLevel_To_Syslog_Priority (level),
	    "%s: %s, system message: %s, object: %s,"
	    " operation: %s, file: %s, line: %d",
	    log_level_description[level],
	    message, strerror (errno), object_description,
	    operation_description, source_file_name,
	    source_file_line_number);
  }
  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr,
	     "%s:\t%s\n"
	     "\t\tsystem message:\t%s\n"
	     "\t\tobject:\t\t%s\n"
	     "\t\toperation:\t%s\n"
	     "\t\tfile:\t\t%s\n"
	     "\t\tline:\t\t%d\n",
	     log_level_description[level],
	     message, strerror (errno), object_description,
	     operation_description, source_file_name,
	     source_file_line_number);
    fflush (local_log_file_ptr);
  }
}				/* Log_Hidden_System_Error() */


void
Log_Buffer (char *file, int line, char *buffer, int buffer_size, char *message)
{
  int i;

  if (local_file_log_is_active) {
    fprintf (local_log_file_ptr, "%s (%s,%d): ", message, file, line);
    for (i = 0; i < buffer_size; i++) {
      if (iscntrl(buffer[i]))
	fprintf (local_log_file_ptr, "[%d]", buffer[i]);
      else
	fputc (buffer[i], local_log_file_ptr);
    }
    fprintf (local_log_file_ptr, "\n");
    fflush (local_log_file_ptr);
  }
}

void 
Close_Log ()
{
  if (syslog_log_is_active) {
    closelog ();
    syslog_log_is_active = FALSE;
  }
  if (local_file_log_is_active) {
    fclose (local_log_file_ptr);
    local_file_log_is_active = FALSE;
  }
  if (user_log_is_active) {
    user_log_is_active = FALSE;
  }
  if (user_fd_log_is_active) {
    user_fd_log_is_active = FALSE;
  }
  if (Debug_Pipe_Is_Open ()) {
    Close_Debug_Pipe ();
  }
  if (inhouse_stderr_log_is_active) {
    inhouse_stderr_log_is_active = FALSE;
  }
}

int 
LogLevel_To_Syslog_Priority (LogLevel level)
{
  switch (level) {
    case LOG_LEVEL_EMERG:
    return LOG_EMERG;
    break;
  case LOG_LEVEL_ALERT:
    return LOG_ALERT;
    break;
  case LOG_LEVEL_CRIT:
    return LOG_CRIT;
    break;
  case LOG_LEVEL_ERR:
    return LOG_ERR;
    break;
  case LOG_LEVEL_WARNING:
    return LOG_WARNING;
    break;
  case LOG_LEVEL_NOTICE:
    return LOG_NOTICE;
    break;
  case LOG_LEVEL_INFO:
    return LOG_INFO;
    break;
  default:
    return LOG_DEBUG;
    break;
  }
}

int 
Write_All (int fd, char *buffer, int n)
{
  int             remaining;
  int             num_written;

  remaining = n;
  while (remaining > 0) {
    num_written = write (fd, buffer, remaining);
    if (num_written == WRITE_ERROR) {
      if (errno == EAGAIN)
	continue;
      return WRITE_ERROR;
    }
    remaining -= num_written;
  }
  return n;
}
