#!/usr/local/bin/expect

source "rcat.conf.tcl"

set report_file [open $reportname w]
puts $report_file "session# filesize #downloads avgtime(sec) avgcps avgbaud"

set totalavgtime 0
set totalavgcps  0
set totalavgbaud 0
set numvalidsessions 0

for {set session_number 1} {$session_number <= $numsessions} \
	{incr session_number} {
    set session_file_name "$datafilebase.$session_number"
    if { ! [file exists $session_file_name] } {
	puts $report_file [format "%8d file does not exist" $session_number]
	continue
    }
    set session_file [open $session_file_name r]
    if {! [eof $session_file]} {
	set total_time 0
	if {-1 == [gets $session_file download_file_size]} {
	    puts $report_file [format "%8d file is empty" $session_number]
	    continue
	}
	set j 0
	while {! [eof $session_file]} {
	    gets $session_file line
	    set num_conversions [scan $line "%d" microsecond_download_time]
	    if {$num_conversions == 1} {
		set download_time [expr $microsecond_download_time / 1000000]
		incr total_time $download_time
		incr j
	    }
	}
	if {$j > 0} {
	    set avgtime [expr $total_time / $j]
	    set avgcps [expr $download_file_size / $avgtime]
	    set avgbaud [expr 8 * $avgcps]
	    puts $report_file [format "%8d %8d %10d %12d %6d %7d" \
		    $session_number \
		    $download_file_size $j $avgtime $avgcps $avgbaud]
	    incr totalavgtime $avgtime
	    incr totalavgcps  $avgcps
	    incr totalavgbaud $avgbaud
	    incr numvalidsessions
	} else {
	    puts $report_file [format "%8d %8d session must have timed out" \
		    $session_number $download_file_size]
	}
    }
    close $session_file
}

puts $report_file "-----------------------------------------------------------"
puts $report_file "                         totals"
puts $report_file "-----------------------------------------------------------"
puts $report_file [format "                             %12d %6d %7d" \
	$totalavgtime $totalavgcps $totalavgbaud]

set overallavgtime [expr $totalavgtime / $numvalidsessions]
set overallavgcps  [expr $totalavgcps  / $numvalidsessions]
set overallavgbaud [expr $totalavgbaud / $numvalidsessions]

puts $report_file "-----------------------------------------------------------"
puts $report_file "                     over all averages"
puts $report_file "-----------------------------------------------------------"
puts $report_file [format "                             %12d %6d %7d" \
	$overallavgtime $overallavgcps $overallavgbaud]

close $report_file
