#!/usr/local/bin/expect

set ERROR 1
set SUCCESS 0

proc kill_process {pid spawn_id} {
    global SUCCESS ERROR out

    set status [catch {exec /bin/kill $pid}]
    if {$status == $ERROR} {
	puts $out "ERROR: could not kill process $pid"
    }
    close
}

proc receive_host_name_prompt {pid spawn_id} {
    global ERROR SUCCESS out
    
    expect {
	timeout {
	    puts $out "ERROR: timed out waiting for Host Name prompt"
	    kill_process $pid $spawn_id
	    return $ERROR
	}
	"Connection refused" {
	    puts $out "ERROR: Connection refused"
	    catch {close}
	    catch {wait}
	    return $ERROR
	}
	"Unknown host*$" {
	    puts $out "ERROR: $expect_out(0,string)"
	    catch {close}
	    catch {wait}
	    return $ERROR
	}
	"Unable to get host name for host*$" {
	    puts $out "ERROR: could not get host name - dns problem?"
	    catch {close}
	    catch {wait}
	    return $ERROR
	}
	"Host Name:*$" {
	    return $SUCCESS
	}
    }
    puts $out "ERROR: inetcis went away at Host Name prompt?"
    wait
    return $ERROR
}

proc receive_baud_rate_prompt {pid spawn_id} {
    global ERROR SUCCESS out

    expect {
	timeout {
	    puts $out "ERROR: timed out waiting for baud rate prompt"
	    kill_process $pid $spawn_id
	    return $ERROR
	}
	" OFF):*$" {
	    return $SUCCESS
	}
    }
    puts $out "ERROR: inetcis went away at baud rate prompt?"
    close
    wait
    return $FAILURE
}

proc receive_user_id_prompt {pid spawn_id} {
    global ERROR SUCCESS out

    expect {
	timeout {
	    puts $out "ERROR: timed out waiting for User ID prompt"
	    kill_process $pid $spawn_id
	    return $ERROR
	}
	"?*ICD*$" {
	    puts $out "ERROR: $expect_out(0,string)"
	    catch {wait}
	    return $ERROR
	}
	"User ID:*$"
    }
    return $SUCCESS
}

proc check_machine {machine_name} {
    global ERROR SUCCESS out

    puts -nonewline $out "$machine_name: "
    set pid [spawn inetcis $machine_name]
    set status [receive_host_name_prompt $pid $spawn_id]
    if {$status == $ERROR} {
	return $ERROR
    }
    send "cisagree\n"
    set status [receive_baud_rate_prompt $pid $spawn_id]
    if {$status == $ERROR} {
	return $ERROR
    }
    send "infinity\n"
    set status [receive_user_id_prompt $pid $spawn_id]
    if {$status == $ERROR} {
	return $ERROR
    }
    kill_process $pid $spawn_id
    wait
    puts $out "OK"
    return $SUCCESS
}

proc check_machine_thrise {machine_name} {
    global ERROR SUCCESS out

    for {set i 1} {$i <= 3} {incr i} {
        set status [check_machine $machine_name]
	if {$status == $ERROR} {
	    return $ERROR
	}
    }
    return $SUCCESS
}

proc check_center {machine_prefix number_of_machines host_to_skip} {
    global ERROR SUCCESS out

    for {set i 1} {$i <= $number_of_machines} {incr i} {
	if {$i == $host_to_skip} {
	    continue
	}
	check_machine_thrise $machine_prefix-gw-$i.compuserve.com
    }
}

#
# Parse command line
#
set mail_data_file 0
for {set i 1} {$i <= $argc} {incr i} {
	if {[lindex $argv [expr $i - 1]] == "-m"} {
		if {$i+1 <= $argc} {
			incr i
			set mail_data_file 1
			set mail_address [lindex $argv [expr $i - 1]]
		}
	}
}

set out_file_name "X25-LINK-STATUS"
set err_file_name "X25-LINK-ERRORS"
set out [open $out_file_name "w"]

check_center "arl" 65 27
check_center "dub" 83 27
# The `0' means that there are no hosts to skip.
check_center "hil" 64 0

close $out

exec grep -i error $out_file_name > $err_file_name

set mail_log_file [open "X25-MAIL-LOG" "w"]
if {$mail_data_file == 1} {
	exec /usr/bin/mail -s "X.25 Link Status" $mail_address < $err_file_name
	puts $mail_log_file "Mailed $err_file_name to $mail_addres"
} else {
	puts $mail_log_file "Mail switch not specified - no mail sent"
}
close $mail_log_file
