#!/usr/local/bin/expect

source "global.tcl"

proc kill_telnet {spawn_id pid} {
    catch {exec kill $pid}
    close
    wait
}

proc verify {results_file machine_prefix start_number stop_number domain_name \
	user_prompt reboot} {
    global spawn_id
    global expect_out
    global xpad_string
    global cud_conf_string
    source "global.tcl"
    
    set timeout 180
    set everything_is_ok 1

    for {set i $start_number} {$i <= $stop_number} {incr i} {
	set machine_name "$machine_prefix$i.$domain_name"
	puts -nonewline $results_file "$machine_name: "
	set telnet_pid [spawn telnet $machine_name $hide_telnet_port]
	expect {
	    timeout {
		puts $results_file "ERROR - expect timed out waiting for login prompt"
		kill_telnet $spawn_id $telnet_pid
		continue
	    }
	    "Unknown host*$" {
		puts $results_file "ERROR - telnet returned error Unknown host"
		wait
		continue
	    }
	    "telnet: Unable to connect to remote host: Connection timed out" {
		puts $results_file "ERROR - telnet connect attempt timed out"
		wait ;# for telnet process (child) to exit
		continue
	    }
	    "login:*$"
	}

	send "$user_name\n"
	expect {
	    timeout {
		puts $results_file "ERROR - expect timed out waiting for password prompt"
		kill_telnet $spawn_id $telnet_pid
		continue
	    }
	    "Password:*$"
	}

	send "$user_password\n"
	expect {
	    timeout {
		puts $results_file "ERROR - expect timed out waiting for shell prompt"
		kill_telnet $spawn_id $telnet_pid
		continue
	    }
	    "Login incorrect*$" {
		puts $results_file "ERROR - wrong user password"
		kill_telnet $spawn_id $telnet_pid
		continue
	    }
	    "$user_prompt"
	}

	send "ls -l /usr/local/adm/daemon/xpad\n"
	expect {
	    "ls*No such file or directory" {
		puts $results_file "ERROR - xpad file does not exist"
		set everything_is_ok 0
	    }
	    timeout {
		puts $resutls_file "ERROR - expect timed out waiting for xpad listing"
		set everything_is_ok 0
	    }
	    -- "-r*xpad\\*"
	}
	if {$everything_is_ok} {
	    set ls_line $expect_out(0,string)     ;# move it into a named var
	    if {[string compare $ls_line $xpad_string] != 0} {
		set everything_is_ok 0
		puts $results_file "ERROR - wrong xpad - this is what is there"
		puts $results_file $ls_line
	    }
	}

	expect $user_prompt
	send "ls -l /etc/cud.conf\n"
	expect {
	    "ls*No such file or directory" {
		puts $results_file "ERROR - cud.conf file does not exist"
		set everything_is_ok 0
	    }
	    timeout {
		puts $results_file "ERROR - expect timed out waiting for cud.conf listing"
		set everything_is_ok 0
	    }
	    -- "-r*cud.conf"
	}
	if {$everything_is_ok} {
	    set ls_line $expect_out(0,string)
	    if {[string compare $ls_line $cud_conf_string] != 0} {
		set everything_is_ok 0
		puts $results_file "ERROR - wrong cud.conf - this is what's there"
		puts $results_file $ls_line
	    }
	}

	expect $user_prompt
	send "ls -l /usr/local/adm/bin/inetcis.login\n"
	expect {
	    "ls*No such file or directory" {
		puts $results_file "ERROR - inetcis.login file does not exist"
		set everything_is_ok 0
	    }
	    timeout {
		puts $results_file "ERROR - expect timed out waiting for inetcis.login listing"
		set everything_is_ok 0
	    }
	    -- "-r*inetcis.login\\*"
	}
	if {$everything_is_ok} {
	    set ls_line $expect_out(0,string)
	    if {[string compare $ls_line $inetcis_login_string] != 0} {
		set everything_is_ok 0
		puts $results_file "ERROR - wrong inetcis.login - this is what's there"
		puts $results_file $ls_line
	    }
	}

	expect $user_prompt
	if {$everything_is_ok && $reboot} {
	    send "su\n"
	    expect "Password:*$"
	    send "$root_password\n"
	    expect {
		"Sorry*$" {
		    puts $results_file "ERROR - wrong root password"
		    kill_telnet $spawn_id $telnet_pid
		    continue
		}
		"$root_prompt"
	    }
	    send "shutdown -r now\n"
	    expect {
		"shutdown: Exec format error" {
		    puts $results_file "ERROR - got Exec format error from shutdown - machine may still reboot"
		    kill_telnet $spawn_id $telnet_pid
		    continue
		}
		"Connection closed by foreign host*$"
	    }
	    puts $results_file "Looks like a successful reboot."
	    wait
	} elseif {$everything_is_ok} {
	    puts $results_file "all files look OK - no reboot"
	    send "exit\n"
	    expect "Connection closed by foreign host.$"
	    wait
	} else {
	    send "exit\n"
	    expect "Connection closed by foreign host.$"
	    wait
	}
    }
}


puts -nonewline "Check machines in (A)rlington, (D)ublin, or (b)oth? "
gets stdin which_machines
puts -nonewline "Reboot these machines if they are OK? "
gets stdin reboot
if {$reboot == "yes" || $reboot == "YES" || $reboot == "Y" \
	|| $reboot == "Yes" || $reboot == "y"} {
    set reboot 1
} else {
    set reboot 0
}
puts -nonewline "Start with which machine number (1-65)? "
gets stdin start_machine
puts -nonewline "Stop at which machine number ($start_machine-65)? "
gets stdin stop_machine
puts -nonewline "Name of in which to store results \[GATEWAY-REBOOT-RESULTS\]? "
gets stdin results_file_name
if {$results_file_name == ""} {
    set results_file_name "GATEWAY-REBOOT-RESULTS"
}

set results_file [open $results_file_name w]

if {$which_machines == "A" || $which_machines == "a" \
	|| $which_machines == "B" || $which_machines == "b"} {
    verify $results_file "arl-gw-" $start_machine $stop_machine $domain_name \
	    $arlington_user_prompt $reboot
}

if {$which_machines == "D" || $which_machines == "d" \
	|| $which_machines == "B" || $which_machines == "b"} {
    verify $results_file "dub-gw-" $start_machine $stop_machine $domain_name \
	    $dublin_user_prompt $reboot
}

close $results_file
    
    